/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd;

import java.io.FileNotFoundException;
import java.io.IOException;

import org.apache.commons.cli.BasicParser;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.OptionBuilder;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.apache.log4j.Logger;
import org.openrdf.repository.RepositoryException;

/**
 * RDDChecker main class, containing the main method for starting the checker.
 * 
 * @author MSchmidt
 */
public class RDDChecker {
	
	private static Logger logger = Logger.getLogger(RDDChecker.class);
	
	public static void main(String [] args) 
	{
		logger.info("Starting up RDD Processor");
		
		Options cmdOptions = cmdLineOptions();
		CommandLineParser parser = new BasicParser();
		
		try {
			logger.info("Parsing command line...");
			CommandLine cmd = parser.parse( cmdOptions, args);

			logger.info("Creating config...");
			Config conf = new Config(cmd);
			
			logger.info("Initializing executor...");
			RDDExecutor rddExec = new RDDExecutor(conf);
			
			logger.info("Running executor...");
			try {
				rddExec.execute();
			} catch (FileNotFoundException e) {
				System.out.println("File cannot be accessed: " + e.getMessage());				
			} catch (IOException e) {
				System.out.println("IOException: " + e.getMessage());								
			} catch (edu.freiburg.dbis.rdd.parser.ParseException e) {
				System.out.println("Parse error in RDD file: " + e.getMessage());				
			} catch (RepositoryException e) {
				System.out.println("Error in Sesame Repository Management: " + e.getMessage());				
			}
		} catch (ParseException e) {
			System.out.println("Illegal program call: " + e.getMessage());
			printHelpText(cmdOptions);
		} catch (IllegalArgumentException e) {
			System.out.println("Illegal program arguments: " + e.getMessage());
			printHelpText(cmdOptions);
		}
		
		logger.debug("RDD Processor finished");
	}
	
	/**
	 * Command line options for the RDD Processor.
	 * 
	 * Example arguments:
	 * <code>-r examples/test/semantics/MixedFeature.rdd -t FOL</code> 
	 */
	protected static Options cmdLineOptions() {

		Options options = new Options();

		// add options		
		Option oHelp = OptionBuilder.create("h");
		oHelp.setDescription("display help");
		oHelp.setRequired(false);
		oHelp.setLongOpt("help");
		oHelp.setValueSeparator('=');
		oHelp.setArgs(0);

		Option oType = OptionBuilder.create("t");
		oType.setDescription("output format type: ASK, SELECT, FOL");
		oType.setRequired(false);
		oType.setLongOpt("type");
		oType.setValueSeparator('=');
		oType.setArgs(1);

		Option oFile = OptionBuilder.create("o");
		oFile.setDescription("output file (if not specified: STDOUT)");
		oFile.setRequired(false);
		oFile.setLongOpt("outputfile");
		oFile.setValueSeparator('=');
		oFile.setArgs(1);

		Option oLimit = OptionBuilder.create("l");
		oLimit.setDescription("number of witnesses to show for constraint violations, where 0 means all (if not specified: 3)");
		oLimit.setRequired(false);
		oLimit.setLongOpt("limit");
		oLimit.setValueSeparator('=');
		oLimit.setArgs(1);

		Option oErrFile = OptionBuilder.create("e");
		oErrFile.setDescription("error file (if not specified: not written)");
		oErrFile.setRequired(false);
		oErrFile.setLongOpt("errorfile");
		oErrFile.setValueSeparator('=');
		oErrFile.setArgs(1);
		
		Option oStatsFile = OptionBuilder.create("s");
		oStatsFile.setDescription("statistics file (if not specified: not written)");
		oStatsFile.setRequired(false);
		oStatsFile.setLongOpt("statsfile");
		oStatsFile.setValueSeparator('=');
		oStatsFile.setArgs(1);		
		
		Option oRdd = OptionBuilder.create("r");
		oRdd.setDescription("RDD input file");
		oRdd.setRequired(true);
		oRdd.setLongOpt("rdd");
		oRdd.setValueSeparator('=');
		oRdd.setArgs(1);
		
		Option oRdf = OptionBuilder.create("d");
		oRdf.setDescription("RDF data input file");
		oRdf.setRequired(false);
		oRdf.setLongOpt("data");
		oRdf.setArgs(1);
		
		Option oMime = OptionBuilder.create("m");
		oMime.setDescription("Mime type of RDF data input file (if not specified, "
				+ "guessed from file ending, default: application/rdf+xml)");
		oMime.setRequired(false);
		oMime.setLongOpt("mimetype");
		oMime.setArgs(1);
		
		options.addOption(oHelp);
		options.addOption(oType);
		options.addOption(oFile);
		options.addOption(oErrFile);
		options.addOption(oLimit);
//		options.addOption(oStatsFile);
		options.addOption(oRdd);
		options.addOption(oRdf);
		options.addOption(oMime);
		
		return options;
	}
	
	/**
	 * Returns the help text describing the command line.
	 */
	protected static void printHelpText(Options cmdOptions) {
		HelpFormatter formatter = new HelpFormatter();
		formatter.printHelp("-r <arg> [-d <arg>] [-t <arg>] [-m <arg>] [-o <arg>] [-e <arg>] [-l <arg>]", cmdOptions);
	}
}