/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.freiburg.dbis.rdd.fol.FOFormula;
import edu.freiburg.dbis.rdd.provenance.ConstraintProvenanceInfo;
import edu.freiburg.dbis.rdd.provenance.ConstraintProvenanceInfo.ConstraintClass;
import edu.freiburg.dbis.rdd.sparql.IRI;

/**
 * Class representing the evaluation result of an RDD, executed over
 * an RDF dataset.
 * 
 * @author MSchmidt
 */
public class RDDEvaluationResult {

	protected List<RDDConstraintEvaluationResult> rddConstraintEvaluationResults;

	public RDDEvaluationResult() {
		rddConstraintEvaluationResults = 
			new ArrayList<RDDConstraintEvaluationResult>();
	}
	
	public void recordNextRDDConstraintEvaluation(ConstraintProvenanceInfo constraint) {
		rddConstraintEvaluationResults.add(
			new RDDConstraintEvaluationResult(constraint));
	}
	
	public void startRecordFOFormulaEvaluation(FOFormula formula, String query) {
		if (rddConstraintEvaluationResults.isEmpty())
			throw new RuntimeException("Cannot start recording FOFormula before recording associated RDD constraint.");

		RDDConstraintEvaluationResult rddConstraintEvaluationResult =
			rddConstraintEvaluationResults.get(rddConstraintEvaluationResults.size()-1);
		rddConstraintEvaluationResult.startRecordFOFormulaEvaluation(formula, query);
	}
	
	public void stopRecordFOFormulaEvaluation(boolean success) {
		if (rddConstraintEvaluationResults.isEmpty())
			throw new RuntimeException("Cannot stop recording FOFormula before start recording FOFormula.");

		RDDConstraintEvaluationResult rddConstraintEvaluationResult =
			rddConstraintEvaluationResults.get(rddConstraintEvaluationResults.size()-1);
		rddConstraintEvaluationResult.stopRecordFOFormulaEvalation(success);
	}
	
	public void registerCounterExample(String counterExampleStr, int total, Integer limit) {
		
		RDDConstraintEvaluationResult rddConstraintEvaluationResult =
				rddConstraintEvaluationResults.get(rddConstraintEvaluationResults.size()-1);
			rddConstraintEvaluationResult.registerCounterExample(counterExampleStr, total, limit);
	}

	public boolean allSucceeded() {
		boolean success = true;
		for (RDDConstraintEvaluationResult rddConstraintEvaluationResult : rddConstraintEvaluationResults) {
			success &= rddConstraintEvaluationResult.evaluatedSuccessfuly();
		}
		return success;
	}
	
	public boolean allRddFailed() {
		boolean allFailed = true;
		for (RDDConstraintEvaluationResult rddConstraintEvaluationResult : rddConstraintEvaluationResults) {
			allFailed &= !rddConstraintEvaluationResult.evaluatedSuccessfuly();
		}
		return allFailed;
		
	}
	
	/**
	 * Return statistics as string
	 */
	public String statsAsString() {
		return new String("Feature not yet implemented");
	}
	
	@Override
	public String toString() {
		return resultAsString(false);
	}
	
	public String resultAsString(boolean showFailuresOnly) {
		StringBuilder buf = new StringBuilder();
		
		String resultStr = allSucceeded() ? "CONSISTENT" : "INCONSISTENT";

		int rddConstraintSuccess = 0;
		int rddConstraintFailure = 0;
		int foConstraintSuccess = 0;
		int foConstraintFailure = 0;

		/**
		 * Calculate global stats
		 */
		Long evaluationTime = Long.valueOf(0);
		for (RDDConstraintEvaluationResult rddConstraintEvaluationResult : rddConstraintEvaluationResults) {

			if (rddConstraintEvaluationResult.evaluatedSuccessfuly())
				rddConstraintSuccess++;
			else 
				rddConstraintFailure++;

			evaluationTime += rddConstraintEvaluationResult.evaluationTime();
			
			for (FOFormulaEvaluationResult foFormulaEvaluationResult : rddConstraintEvaluationResult.getFOFormulaEvaluationResults()) {
				if (foFormulaEvaluationResult.evaluatedSuccessfuly())
					foConstraintSuccess++;
				else
					foConstraintFailure++;
			}
		}

		/**
		 * Get detailed report
		 */
		StringBuilder details = new StringBuilder();
		StructuredRDDConstraintEvaluationResult structuredResult =
				new StructuredRDDConstraintEvaluationResult(rddConstraintEvaluationResults);
		details.append(structuredResult.toString(showFailuresOnly));
		
		buf.append("############## The RDF document is " + resultStr + " with the RDD\n");
		buf.append("############## Evaluation summary" + "\n");
		buf.append("############ RDD constraints success/failure: " + rddConstraintSuccess + "/" + rddConstraintFailure + "\n");
		buf.append("############ FOL constraints success/failure: " + foConstraintSuccess + "/" + foConstraintFailure + "\n");
		buf.append("############ Evaluation time: " + evaluationTime + "ms\n");
		buf.append("\n");
		buf.append("\n");
		
		if (showFailuresOnly) {
			buf.append("############## Evaluation failure report (skipping succeeded constraints)\n");						
		} else {
			buf.append("############## Complete evaluation report\n");			
		}
		buf.append(details.toString());
		
		return buf.toString();
	}

	public static class RDDConstraintEvaluationResult {
		ConstraintProvenanceInfo provenanceInfo;
		List<FOFormulaEvaluationResult> foFormulaEvaluationResults;
		
		public RDDConstraintEvaluationResult(ConstraintProvenanceInfo provenanceInfo) {
			this.provenanceInfo = provenanceInfo;
			foFormulaEvaluationResults = new ArrayList<FOFormulaEvaluationResult>();
		}
		
		public void startRecordFOFormulaEvaluation(FOFormula formula, String query) {
			foFormulaEvaluationResults.add(new FOFormulaEvaluationResult(formula, query));
		}
		
		public void stopRecordFOFormulaEvalation(boolean success) {
			if (foFormulaEvaluationResults==null || foFormulaEvaluationResults.isEmpty())
				throw new RuntimeException("Illegal use of RDDEvaluationResult: cannot stop recording prior to starting.");
			FOFormulaEvaluationResult curEvaluationResult = 
				foFormulaEvaluationResults.get(foFormulaEvaluationResults.size()-1);
			curEvaluationResult.stopEvaluation(success);
		}
		
		public void registerCounterExample(String counterExampleStr, int total, Integer limit) {
			if (foFormulaEvaluationResults==null || foFormulaEvaluationResults.isEmpty())
				throw new RuntimeException("Illegal use of RDDEvaluationResult: cannot register counterexample prior to starting.");
			FOFormulaEvaluationResult curEvaluationResult = 
				foFormulaEvaluationResults.get(foFormulaEvaluationResults.size()-1);
			curEvaluationResult.registerCounterExample(counterExampleStr, total, limit);
		}

		boolean evaluatedSuccessfuly() {
			boolean success = true;
			for (FOFormulaEvaluationResult foFormulaEvaluationResult : foFormulaEvaluationResults) {
				success &= foFormulaEvaluationResult.evaluatedSuccessfuly();
			}
			return success;
		}
		
		public List<FOFormulaEvaluationResult> getFOFormulaEvaluationResults() {
			return foFormulaEvaluationResults;
		}

		public Long evaluationTime() {
			Long evaluationTime = Long.valueOf(0);
			for (FOFormulaEvaluationResult foFormulaEvaluationResult : foFormulaEvaluationResults) {
				evaluationTime += foFormulaEvaluationResult.evaluationTime();
			}
			return evaluationTime;
		}

		public ConstraintProvenanceInfo getProvenanceInfo() {
			return provenanceInfo;
		}
		
		public String toString(String id) {
			StringBuilder buf = new StringBuilder();
			String resultStr = evaluatedSuccessfuly() ? "success" : "failure";
			buf.append("######## (" + id + ") " + provenanceInfo.constraintClassAsString() + " constraint" + "\n");

			
			int i=0;
			int folConstraintSuccess = 0;
			int folConstraintFailure = 0;
			Long evaluationTime = Long.valueOf(0);
			StringBuilder details = new StringBuilder();
			for (FOFormulaEvaluationResult foFormulaEvaluationResult : foFormulaEvaluationResults) {
				if (foFormulaEvaluationResult.evaluatedSuccessfuly())
					folConstraintSuccess++;
				else
					folConstraintFailure++;
				evaluationTime += foFormulaEvaluationResult.evaluationTime();
				details.append(foFormulaEvaluationResult.toString(id + "." +  i++));
			}
			String constraintRep = provenanceInfo.getConstraintRep();
			if (constraintRep!=null && !constraintRep.isEmpty()) {
				buf.append("###### Textual representation:\n" + provenanceInfo.getConstraintRep() + "\n");
			}
			buf.append("###### Evaluation time: " + evaluationTime + "ms\n");
			buf.append("###### Result: " + resultStr + "\n");
			buf.append("###### Implied FOL constraints: " + (folConstraintSuccess+folConstraintFailure) + 
					" total (" + folConstraintSuccess + " succeeded, " + folConstraintFailure + " failed)\n");
			buf.append(details.toString());
			
			return buf.toString();
		}
	}
	
	public static class FOFormulaEvaluationResult {
		FOFormula formula;
		String query;
		String counterExample;
		Integer counterExamplesTotal;
		Integer counterExamplesLimit;
		Long startTime;
		Long endTime;
		Boolean success;
		
		public FOFormulaEvaluationResult(FOFormula formula, String query) {
			this.formula = formula;
			this.query = query;
			startTime = System.currentTimeMillis();
		}

		public Long evaluationTime() {
			return endTime-startTime;
		}

		public void stopEvaluation(boolean success) {
			endTime = System.currentTimeMillis();
			this.success = success;
		}

		public void registerCounterExample(String counterExample, int total, Integer limit) {
			this.counterExample = counterExample;
			this.counterExamplesTotal = total;
			this.counterExamplesLimit = limit==null ? -1 : limit;
		}

		public boolean evaluatedSuccessfuly() {
			return success;
		}
		
		public Long durationMS() {
			return endTime-startTime;
		}
		
		public String toString(String id) {
			StringBuilder buf = new StringBuilder();
			String resultStr = evaluatedSuccessfuly() ? "success" : "failure";
			buf.append("#### (" + id + ") Implied FOL constraint\n");
			buf.append("## FO Formula: " + "\n");
			buf.append(formula.toRIF() + "\n");
			buf.append("## Query: " + "\n");
			buf.append(query + "\n");
			buf.append("## Evaluation time: " + durationMS() + "ms\n");
			buf.append("## Result: " + resultStr + "\n");
			if (counterExample!=null) {
				buf.append("## Counterexamples (showing " + counterExamplesTotal + " counterexamples");
				if (counterExamplesLimit>=counterExamplesTotal)
					buf.append(" -> limit reached, there may be more");
				buf.append("): \n");
				buf.append(counterExample + "\n");
			}
			return buf.toString();
		}
	}
	
	/**
	 * A container allowing structured access to an
	 * {@link RDDConstraintEvaluationResult}.
	 * 
	 * @author MSchmidt
	 *
	 */
	public static class StructuredRDDConstraintEvaluationResult {

		// section globals
		public RDDConstraintEvaluationResult classesCWAConstraint;
		public RDDConstraintEvaluationResult propertiesCWAConstraint;
		
		// class globals
		public Map<IRI, RDDConstraintEvaluationResult> classCWAConstraintMap;
		public Map<IRI, RDDConstraintEvaluationResult> classSingletonConstraintMap;

		// class key + props
		public Map<IRI, List<RDDConstraintEvaluationResult>> classKeyConstraintsMap;
		public Map<IRI, List<RDDConstraintEvaluationResult>> qualifiedPropConstraintsMap;

		// global props
		public List<RDDConstraintEvaluationResult> unqualifiedPropConstraints;
		
		public StructuredRDDConstraintEvaluationResult(
				
			List<RDDConstraintEvaluationResult> rddConstraintEvaluationResults) {
			
			classCWAConstraintMap = new HashMap<IRI, RDDConstraintEvaluationResult>();
			classSingletonConstraintMap = new HashMap<IRI, RDDConstraintEvaluationResult>();
			classKeyConstraintsMap = new HashMap<IRI, List<RDDConstraintEvaluationResult>>();
			qualifiedPropConstraintsMap = new HashMap<IRI, List<RDDConstraintEvaluationResult>>();
			unqualifiedPropConstraints = new ArrayList<RDDConstraintEvaluationResult>();
			
			for (RDDConstraintEvaluationResult res : rddConstraintEvaluationResults) {
				
				ConstraintProvenanceInfo pi = res.getProvenanceInfo();
				ConstraintClass constraintClass = pi.getConstraintClass();
				IRI referredClass = pi.getReferredClass();
				
				switch (constraintClass) {
				case CLASSES_CWA_CONSTRAINT:
				{
					if (classesCWAConstraint!=null)
						throw new RuntimeException("Duplicate CLASSES CWA constraint.");
					classesCWAConstraint = res;
					break;
				}
				case CLASS_CWA_CONSTRAINT:
				{
					if (classCWAConstraintMap.containsKey(referredClass))
						throw new RuntimeException("Duplicate CLASS CWA constraint for class '" +  referredClass + "'.");
					classCWAConstraintMap.put(referredClass, res);
					break;
				}
				case CLASS_SINGLETON_CONSTRAINT:
				{
					if (classSingletonConstraintMap.containsKey(referredClass))
						throw new RuntimeException("Duplicate CLASS SINGLETON constraint for class '" +  referredClass + "'.");
					classSingletonConstraintMap.put(referredClass, res);
					break;
				}
				case CLASS_KEY_CONSTRAINT:
				{
					List<RDDConstraintEvaluationResult> l = classKeyConstraintsMap.get(referredClass);
					if (l==null)
						l = new ArrayList<RDDConstraintEvaluationResult>();
					l.add(res);
					classKeyConstraintsMap.put(referredClass, l);
					break;
				}
				case QUALIFIED_PROPERTY_CONSTRAINT:
				{
					List<RDDConstraintEvaluationResult> l = qualifiedPropConstraintsMap.get(referredClass);
					if (l==null)
						l = new ArrayList<RDDConstraintEvaluationResult>();
					l.add(res);
					qualifiedPropConstraintsMap.put(referredClass, l);
					break;
				}
				case PROPERTIES_CWA_CONSTRAINT:
				{
					if (propertiesCWAConstraint!=null)
						throw new RuntimeException("Duplicate PROPERTIES CWA constraint.");
					propertiesCWAConstraint = res;
					break;
				}
				case UNQUALIFIED_PROPERTY_CONSTRAINT:
				{
					unqualifiedPropConstraints.add(res);
					break;
				}
				default:
					throw new RuntimeException("Unknown constraint class '" + constraintClass + "'.");
				}
			}
		}
		
		public String toString(boolean showFailuresOnly) {
			
			StringBuilder buf = new StringBuilder();
			
			Set<IRI> allOccurringClasses = new HashSet<IRI>();
			allOccurringClasses.addAll(classCWAConstraintMap.keySet());
			allOccurringClasses.addAll(classSingletonConstraintMap.keySet());
			allOccurringClasses.addAll(classKeyConstraintsMap.keySet());
			allOccurringClasses.addAll(qualifiedPropConstraintsMap.keySet());
			
			int classCount = 0;
			buf.append("############ (0) Class specific constraints\n");
			for (IRI clazz : allOccurringClasses) {

				String classId = "0." + classCount++;
				
				RDDConstraintEvaluationResult classCWAConstraint = classCWAConstraintMap.get(clazz);
				RDDConstraintEvaluationResult classSingletonConstraint = classSingletonConstraintMap.get(clazz);
				List<RDDConstraintEvaluationResult> classKeyConstraints = classKeyConstraintsMap.get(clazz);
				List<RDDConstraintEvaluationResult> qualifiedPropConstraints = qualifiedPropConstraintsMap.get(clazz);
				
				// only print information for class if there is at least one failure
				if (showFailuresOnly) {
					Set<RDDConstraintEvaluationResult> all = new HashSet<RDDConstraintEvaluationResult>();
					
					if (classCWAConstraint!=null)
						all.add(classCWAConstraint);
					if (classSingletonConstraint!=null)
						all.add(classSingletonConstraint);
					if (classKeyConstraints!=null)
						all.addAll(classKeyConstraints);
					if (qualifiedPropConstraints!=null)
						all.addAll(qualifiedPropConstraints);
					
					if (!containsFailedConstraints(all))
						continue;
				}
				
				buf.append("########## (" + classId + ") Constraints for class " + clazz + "\n");

				int constraintCount = 0; // class-internal counter for constraints
				
				// first: class CWA constraint
				if (classCWAConstraint!=null) {
					if (!showFailuresOnly || !classCWAConstraint.evaluatedSuccessfuly()) {
						buf.append(classCWAConstraint.toString(classId + "." + constraintCount));
						buf.append("\n");
					}
					constraintCount++;
				}
				
				// second: class SINGLETON constraint
				if (classSingletonConstraint!=null) {
					if (!showFailuresOnly || !classSingletonConstraint.evaluatedSuccessfuly()) {
						buf.append(classSingletonConstraint.toString(classId + "." + constraintCount));
						buf.append("\n");
					}
					constraintCount++;
				}
				
				// third: class KEY constraints
				if (classKeyConstraints!=null) {
					for (int i=0; i<classKeyConstraints.size(); i++) {
						RDDConstraintEvaluationResult res = classKeyConstraints.get(i);
						if (!showFailuresOnly || !res.evaluatedSuccessfuly()) {
							buf.append(res.toString(classId + "." + constraintCount));
							buf.append("\n");
						}
						constraintCount++;
					}
				}
				
				// fourth: class PROPERTY constraints
				if (qualifiedPropConstraints!=null) {
					for (int i=0; i<qualifiedPropConstraints.size(); i++) {
						RDDConstraintEvaluationResult res = qualifiedPropConstraints.get(i);
						if (!showFailuresOnly || !res.evaluatedSuccessfuly()) {
							buf.append(res.toString(classId + "." + constraintCount));
							buf.append("\n");
						}
						constraintCount++;
					}
				}
				
				buf.append("\n");
			}

			// global props
			buf.append("\n");
			buf.append("############ (1) Unqualified property constraints\n");
			if (unqualifiedPropConstraints.isEmpty()) {
				buf.append("(no unqualified property constraints specified)\n");
			} else {
				int constraintCount=0;
				for (RDDConstraintEvaluationResult res : unqualifiedPropConstraints) {
					if (!showFailuresOnly || !res.evaluatedSuccessfuly()) {
						buf.append(res.toString("1." + constraintCount));
						buf.append("\n");
						constraintCount++;
					}
				}
			}
			
			buf.append("\n");
			buf.append("\n");
			buf.append("############ (2) CLASSES CWA constraint\n");
			if (classesCWAConstraint!=null) {
				if (!showFailuresOnly || !classesCWAConstraint.evaluatedSuccessfuly()) {
					buf.append(classesCWAConstraint.toString("2.0"));
					buf.append("\n");
				}
			} else {
				buf.append("(no CLASSES CWA constraint specified)\n");
			}
			
			buf.append("\n");
			buf.append("\n");
			buf.append("############ (3) PROPERTIES CWA constraint\n");
			if (propertiesCWAConstraint!=null) {
				if (!showFailuresOnly || !propertiesCWAConstraint.evaluatedSuccessfuly()) {
					buf.append(propertiesCWAConstraint.toString("3.0"));
					buf.append("\n");
				}
			} else {
				buf.append("(no PROPERTIES CWA constraint specified)");
				buf.append("\n");
			}
			
			return buf.toString();
		}

		private boolean containsFailedConstraints(Collection<RDDConstraintEvaluationResult> evaluationResults) {
			
			for (RDDConstraintEvaluationResult evaluationResult : evaluationResults) {
				if (!evaluationResult.evaluatedSuccessfuly())
					return true;
			}
			
			return false;
		}
	}		
}