/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.openrdf.model.Value;
import org.openrdf.repository.RepositoryException;
import org.openrdf.rio.RDFFormat;

import edu.freiburg.dbis.rdd.Config.OutputType;
import edu.freiburg.dbis.rdd.fol.FOFormula;
import edu.freiburg.dbis.rdd.model.RDD;
import edu.freiburg.dbis.rdd.parser.ParseException;
import edu.freiburg.dbis.rdd.parser.RDDParser;
import edu.freiburg.dbis.rdd.provenance.ConstraintProvenanceInfo;
import edu.freiburg.dbis.rdd.sparql.DataManager;

/**
 * Main execution logics for RDDs, interpreting the user command line configuration.
 * 
 * @author MSchmidt
 */
public class RDDExecutor {

	private static Logger logger = Logger.getLogger(RDDExecutor.class);
	
	Config conf;
	
	/**
	 * Constructor, initializing the executor with a given configuration.
	 */
	public RDDExecutor(Config conf) {
		this.conf = conf;
	}
	
	/**
	 * Main method starting the RDD executor with
	 * its associated configuration.
	 * @throws FileNotFoundException 
	 * @throws ParseException 
	 * @throws RepositoryException 
	 */
	public void execute() 
	throws FileNotFoundException, ParseException, IOException, RepositoryException {
		
		logger.info("Setting up parser.");
		FileInputStream fis = new FileInputStream(conf.getRddFile());
	    RDDParser parser = new RDDParser(fis);
	    
		logger.info("Parsing RDD.");
	    RDD rdd = parser.rdd();

		Set<FOFormula> formulas = rdd.toFOL();

	    if (!conf.isEvaluationMode()) {
	    	outputTranslationResult(formulas, conf.getOutputType(), conf.getLimit(), conf.getOutputFile());
	    } else {
	    	StringBuilder outputBuffer = new StringBuilder();
	    	
	    	OutputType type = conf.getOutputType();
	    	Boolean askMode = null;
	    	
	    	switch (type) {
	    	case ASK:
	    		askMode = true;
	    		break;
	    	case SELECT:
	    		askMode = false;
	    		break;
	    	default:
	    		throw new RuntimeException("Illegal configuration: "
	    				+ "output type must be ASK or SELECT when database file specified.");
	    	}
	    	
	    	RDDEvaluationResult er =
	    		evalResult(formulas, conf.getDatabaseFile(), conf.getDatabaseFileFormat(),
	    					askMode, conf.getLimit());
	    	outputBuffer.append(er.toString());
			
			if (conf.getOutputFile()!=null) {
				System.out.println("See specified output file for full report.");
			}
			
			/**
			 * Write stats, if requested.
			 */
			if (conf.getStatsFile()!=null) {
				outputBuffer.append(er.statsAsString());
				System.out.println("See specified stats file for full report.");
			}
			
			/**
			 * Write error log, if requested.
			 */
			if (conf.getErrorFile()!=null) {
				writeStringToOutput(er.resultAsString(true), conf.getErrorFile());
				System.out.println("See specified error file for error report.");
			}
			
			// write result
			writeStringToOutput(outputBuffer.toString(), conf.getOutputFile());
	    }
	}

	/**
	 * Write the evaluation result to the output.
	 * @throws IOException 
	 */
	public static void outputTranslationResult(Set<FOFormula> formulas, OutputType type, Integer limit, File outputFile)
	throws IOException {

		StringBuilder outputBuffer = new StringBuilder();
		
		// we put the constraints in the order according to their provenance
		Map<ConstraintProvenanceInfo, List<FOFormula>> orderedConstraints = 
				orderByProvenance(formulas);
		
		for (ConstraintProvenanceInfo provenanceInfo : orderedConstraints.keySet()) {

			outputBuffer.append(provenanceInfo.toString() + "\n");
			for (FOFormula formula : orderedConstraints.get(provenanceInfo)) {
				// write provenance information, if requested

				// write output information
				String output = null;
				switch (type) {
				case FOL:
					output = formula.toRIF();
					break;
				case ASK:
					output = formula.toFullSPARQL(true, null); // no limit ->
																// ASK
					break;
				case SELECT:
					output = formula.toFullSPARQL(false, limit);
					break;
				default:
					throw new RuntimeException("Illegal output type: " + type);
				}
				outputBuffer.append(output + "\n");
			}

			// write trailing linebreak (pretty print)
			outputBuffer.append("\n");
		}

		writeStringToOutput(outputBuffer.toString(), outputFile);
		if (outputFile != null)
			System.out.println("See specified output file for full report.");
	}

	/**
	 * Evaluate the result over an RDF database.
	 */
	public static RDDEvaluationResult evalResult(Set<FOFormula> formulas, 
			File dataFile, RDFFormat fileFormat, boolean askMode)
	throws IOException, RepositoryException {
		return evalResult(formulas, dataFile, fileFormat, askMode, 1);
	}

	/**
	 * Evaluate the result over an RDF database.
	 */
	public static RDDEvaluationResult evalResult(Set<FOFormula> formulas, 
			File dataFile, RDFFormat fileFormat, boolean askMode, Integer limit)
	throws IOException, RepositoryException {
		
		RDDEvaluationResult er = new RDDEvaluationResult();
		
		// we put the constraints in the order according to their provenance
		Map<ConstraintProvenanceInfo, List<FOFormula>> orderedConstraints = 
				orderByProvenance(formulas);
		
		DataManager dm = null;
		
		try {
			System.out.println("> Loading data file (this may take a while...)");
			dm = new DataManager(dataFile, fileFormat);		
			System.out.println("> Starting Evaluation");
			System.out.println("> Number of RDD constraints: " + orderedConstraints.keySet().size());
			
			int rddConstraintCtr = 0;
			int nrOfFOConstraints = orderedConstraints.keySet().size();
			for (ConstraintProvenanceInfo provenanceInfo : orderedConstraints.keySet()) {
	
				System.out.println("> Evaluating RDD constraint #" + ++rddConstraintCtr + "/" + nrOfFOConstraints + " (" + orderedConstraints.get(provenanceInfo).size() + " FO constraints)...");
				er.recordNextRDDConstraintEvaluation(provenanceInfo);
				for (FOFormula formula : orderedConstraints.get(provenanceInfo)) {
					if (askMode) {
						String askQuery = formula.toFullSPARQL(true,null); // ASK mode -> no LIMIT 
						er.startRecordFOFormulaEvaluation(formula, askQuery);
						boolean res = dm.sparqlAsk(askQuery);
						er.stopRecordFOFormulaEvaluation(!res);
					} else {
						String selectQuery = formula.toFullSPARQL(false,limit);
						er.startRecordFOFormulaEvaluation(formula, selectQuery);
						List<Map<String,Value>> qr = dm.sparqlSelect(selectQuery);
						if (!qr.isEmpty()) {
							
							StringBuilder counterExampleStr = new StringBuilder();
							for (int i=0; i<qr.size(); i++) {
								Map<String,Value> counterExample = qr.get(i);
								if (i>0)
									counterExampleStr.append("\n");
								
								int j=0;
								for (String key : counterExample.keySet()) {
									if (j++>0)
										counterExampleStr.append(", ");
									counterExampleStr.append(key);
									counterExampleStr.append(" -> ");
									counterExampleStr.append(counterExample.get(key).toString());
								}
								
							}
							er.registerCounterExample(counterExampleStr.toString(), qr.size(), limit);
							er.stopRecordFOFormulaEvaluation(false);
						} else {
							er.stopRecordFOFormulaEvaluation(true);
						}
					}
				}
	
			}
			System.out.println("> All done!");
			System.out.println();
			
		} catch (Exception e) {
			logger.error(e.getMessage());
		} finally {		
			dm.shutdown();
		}
		
		return er;
	}
	
	/**
	 * Groups the constraint by their provenance in the specified RDD
	 * 
	 * @param formulas
	 * @return
	 */
	private static Map<ConstraintProvenanceInfo, List<FOFormula>> 
		orderByProvenance(Set<FOFormula> formulas) {
		
		Map<ConstraintProvenanceInfo, List<FOFormula>> ret = 
			new HashMap<ConstraintProvenanceInfo, List<FOFormula>>();
		for (FOFormula formula : formulas) {
			ConstraintProvenanceInfo cpi = formula.getProvenanceInfo();
			
			List<FOFormula> associatedConstraints = ret.get(cpi);
			if (associatedConstraints==null) {
				associatedConstraints = new LinkedList<FOFormula>();
				ret.put(cpi,associatedConstraints);
			}
			associatedConstraints.add(formula);
		}

		return ret;
	}

	/**
	 * Writes the specified string to the output. If no file is given,
	 * STDOUT will be used
	 * 
	 * @param f
	 * @return
	 */
	static void writeStringToOutput(String s, File f) {
		
		OutputStream os = null;
		try {
			if (f!=null) {
				os = new FileOutputStream(f);
				os.write(strToUTFByteArr(s));
			} else {
				System.out.println(s);
			}
		} catch (Exception e) {
			throw new RuntimeException("Could not open selected output stream: " + e.getMessage());
		} finally {
			if (os!=null)
				try {
					os.close();
				} catch (IOException e) {
				}
		}
	}

	/**
	 * Converts a string to a UTF8 byte[] array
	 */
	static byte[] strToUTFByteArr(String s) {
		return s.getBytes(Charset.forName("UTF-8"));
	}
}