/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd.fol;

import java.util.LinkedList;
import java.util.List;

/**
 * First-order logic "AND" expression
 * 
 * @author MSchmidt
 */
public class FOAnd extends FOFormula {
	protected List<FOFormula> conjunction;
	
	public FOAnd(List<FOFormula> conjunction) {
		this.conjunction = conjunction;
	}
	
	public FOAnd(FOFormula c1, FOFormula c2) {
		conjunction = new LinkedList<FOFormula>();
		conjunction.add(c1);
		conjunction.add(c2);
	}

	@Override
	public String toRIF() {
		StringBuffer buf = new StringBuffer();
		buf.append("And(");
		int i=0;
		for (FOFormula formula : conjunction) {
			if (i++>0) buf.append(" ");
			buf.append(formula.toRIF());
		}
		buf.append(")");
		
		return buf.toString();
	}

	@Override
	public String toSPARQL(boolean ask) {
		List<FOFormula> tripleConjuncts = getTripleConjuncts();
		List<FOFormula> filterConjuncts = getFilterConjuncts();
		
		// triples are mapped to a UNION of triple patterns
		StringBuilder buf = new StringBuilder();
		for (FOFormula tripleConjunct : tripleConjuncts)
			buf.append(tripleConjunct.toSPARQL(ask) + " .\n");
		
		// conditions are mapped to a FILTER expression
		if (!filterConjuncts.isEmpty()) {
			if (!tripleConjuncts.isEmpty())
				buf.append("FILTER (");
			// else: there are no triples, but only datatype restriction (which include a filter anyway)
			
			for (int i=0; i<filterConjuncts.size(); i++) {
				if (i>0) buf.append(" && ");
				buf.append(filterConjuncts.get(i).toSPARQL(ask));
			}
			if (!tripleConjuncts.isEmpty())
				buf.append(")\n");
			// else: there are no triples, but only datatype restriction (which include a filter anyway)
		}		
		return buf.toString();
	}
	
	public List<FOFormula> getTripleConjuncts() {
		List<FOFormula> tripleConjuncts = new LinkedList<FOFormula>();
		for (FOFormula conjunct : conjunction) {
			if (conjunct instanceof FOTriple) {
				tripleConjuncts.add(conjunct);
			} else if ((conjunct instanceof FOEquals) || 
					(conjunct instanceof FONot)) {
				// ignore
			} else if ((conjunct instanceof FORangeType)) {
				// ignore
			} else if ((conjunct instanceof FODatatype)) {
				// ignore
			} else 
				throw new RuntimeException(
					"Case " + conjunct.getClass().getSimpleName() + 
					" not implemented. Aborting in conjunction.");
		}
		
		return tripleConjuncts;
	}

	public List<FOFormula> getFilterConjuncts() {
		List<FOFormula> filterConjuncts = new LinkedList<FOFormula>();
		for (FOFormula conjunct : conjunction) {
			if (conjunct instanceof FOTriple) {
				// ignore
			} else if ((conjunct instanceof FOEquals) || 
					(conjunct instanceof FONot) || 
					(conjunct instanceof FORangeType) ||
					(conjunct instanceof FODatatype)) {
				filterConjuncts.add(conjunct);
			} else {
				throw new RuntimeException(
					"Case " + conjunct.getClass().getSimpleName() + 
					" not implemented. Aborting in conjunction.");
			}
		}
		
		return filterConjuncts;
	}
}