/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd.fol;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import edu.freiburg.dbis.rdd.provenance.ConstraintProvenanceInfo;
import edu.freiburg.dbis.rdd.sparql.SPARQLVariable;


/**
 * 
 * http://www.w3.org/2005/rules/wiki/SWC
 * 
 * Forall ?x ?y ?z (?x[ex:uncleOf -> ?z] :- 
     And(?x[ex:brotherOf -> ?y] ?y[ex:parentOf -> ?z]))
     
   Exists ?x (And(ex:p(?x) ?x[rdf:first -> ex:a] ?x[rdf:rest -> rdf:nil])
 * @author MSchmidt
 *
 */
public abstract class FOFormula {

	protected ConstraintProvenanceInfo provenanceInfo;
	
	public abstract String toRIF();

	/**
	 * Called to construct a full SPARQL query, i.e. including the surrounding
	 * ASK or SELECT clause (depending on the parameter). May be overridden
	 * by the implementing subclass, e.g. to specify the exact list of variables
	 * in SELECT (default is *, i.e. return all variables).
	 */
	public String toFullSPARQL(boolean ask, Integer limit) {
		StringBuffer buf = new StringBuffer();
		buf.append(FOFormula.sparqlQueryPrefix(ask,null));
		buf.append(toSPARQL(ask));
		buf.append(FOFormula.sparqlQuerySuffix(limit));
		return buf.toString();
	}

	public abstract String toSPARQL(boolean ask);
	
	public ConstraintProvenanceInfo getProvenanceInfo() {
		return provenanceInfo;
	}
	
	public void setProvenanceInfo(ConstraintProvenanceInfo provenanceInfo) {
		this.provenanceInfo = provenanceInfo;
	}
	
	/**
	 * Define equality over the fullUri, which defines
	 * the URI uniquely.
	 */
	@Override
	public int hashCode() {
		return toRIF().hashCode();
	}
	
	/**
	 * Define equality over the serialization, for now.
	 */
	@Override
	public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (obj == this)
            return true;
        if (!(obj instanceof FOFormula))
            return false;

        FOFormula objAsFormula = (FOFormula)obj;
        return this.hashCode()==obj.hashCode() && 
        		((this.provenanceInfo==null || objAsFormula.provenanceInfo==null) || 
        		(this.provenanceInfo!=null && objAsFormula.getProvenanceInfo()!=null &&
	        		this.provenanceInfo.getId()==objAsFormula.getProvenanceInfo().getId()));
	}
	
	/**
	 * Removes the provenance info and puts the formulas into a set, thus removing
	 * formulas that only differ in provenance information.
	 */
	public static Set<FOFormula> minimizeByProvenanceRemoval(Set<FOFormula> formulas) {
		Set<FOFormula> newSet = new HashSet<FOFormula>();
		for (FOFormula formula : formulas) {
			formula.setProvenanceInfo(null);
			newSet.add(formula);
		}
		
		return newSet;
	}
	
	public static String sparqlQueryPrefix(boolean ask, List<SPARQLVariable> vars) {

		StringBuilder buf = new StringBuilder();
		if (ask) {
			buf.append("ASK { \n");
		} else {
			buf.append("SELECT ");
			if (vars!=null && !vars.isEmpty()) {
				for (SPARQLVariable var : vars) {
					buf.append(var.toSPARQL() + " ");
				}
			} else {
				buf.append("* "); // select everything
			}
			buf.append("{ \n");
		}

		return buf.toString();
	}
	
	public static String sparqlQuerySuffix(Integer limit) {
		StringBuilder buf = new StringBuilder();
		buf.append("}");
		if (limit!=null && limit>0) {
			buf.append(" LIMIT " + limit);
		}
		return buf.toString();
	}
}
