/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd.fol;

import edu.freiburg.dbis.rdd.sparql.SPARQLVariableFactory;

/**
 * Representation of FO implication construct
 * 
 * @author MSchmidt
 */
public class FOImplies extends FOFormula {

	public static enum ConclusionType {
		TRIPLE_PATTERNS_ONLY,
		EQUALITIES_ONLY,
		MIXED
	}
	
	protected FOFormula premise;
	protected FOFormula conclusion;
	
	public FOImplies(FOFormula premise, FOFormula conclusion) {
		this.premise = premise;
		this.conclusion = conclusion;
	}
	
	@Override
	public String toRIF() {
		StringBuffer buf = new StringBuffer();
		buf.append(conclusion.toRIF());
		buf.append(" :- ");
		buf.append(premise.toRIF());
		
		return buf.toString();
	}
	
	@Override
	public String toSPARQL(boolean ask) {
		StringBuilder buf = new StringBuilder();
		
		// special handling: if the premise is only something like "IRI(?o)",
		// then we need to generate the universe of all values via a UNION query
		if (premise instanceof FORangeType) {
			FORangeType rt = (FORangeType)premise;
			
			if (rt.elem.var!=null) {
				SPARQLVariableFactory vf = SPARQLVariableFactory.getInstance();
				buf.append(" { " + rt.elem.var.toSPARQL() + " " + vf.freshDummyVariable().toSPARQL() + " " + vf.freshDummyVariable() + "}");
				buf.append(" UNION ");
				buf.append("{ " + vf.freshDummyVariable() + " " + rt.elem.var.toSPARQL() + " " + vf.freshDummyVariable() + " }");
				buf.append(" UNION ");
				buf.append("{ " + vf.freshDummyVariable() + " " + vf.freshDummyVariable() + " " + rt.elem.var.toSPARQL() + " } ");
			}
			buf.append(" FILTER ");
			buf.append(premise.toSPARQL(ask));
		} else
			buf.append(premise.toSPARQL(ask));
		
		if (conclusion instanceof FOFalse) {
			// nothing to do
		} else if (conclusion instanceof FOTrue) {
			throw new RuntimeException("Case not implemented, something's wrong here.");
		} else {
		
			// determine the conclusion type
			boolean containsTriplePattern = containsTriplePattern(conclusion);
	
			if (containsTriplePattern) {
				// negation is done here, at top-level
				buf.append(" FILTER NOT EXISTS { ");
				buf.append(conclusion.toSPARQL(ask));
				buf.append(" }");
			} else {
				buf.append(" FILTER (!(");
				buf.append(conclusion.toSPARQL(ask));
				buf.append(" ))");
			}
		}
		
		return buf.toString();
	}

	/**
	 * Checks whether a (possibly nested) FO formula is or contains
	 * a triple pattern 
	 */
	protected boolean containsTriplePattern(FOFormula formula) {
		
		// note: we have one recursion level by design of the formulas,
		//       which makes things much easier
		if (formula instanceof FOExists) {
			FOExists existsFormula = (FOExists)formula;
			return containsTriplePattern(existsFormula.inner);
		} else if (formula instanceof FOAnd) {
			FOAnd andFormula = (FOAnd)formula;
			return !(andFormula.getTripleConjuncts().isEmpty());
		} else if (formula instanceof FOOr) {
			FOOr orFormula = (FOOr)formula;
			return !(orFormula.getTripleDisjuncts().isEmpty());
		} else if (formula instanceof FOTriple) {
			return true;
		} else if (formula instanceof FORangeType) {
			return false;
		} else if (formula instanceof FOEquals ||
					formula instanceof FONot) {
			return false;
		}
		
		throw new RuntimeException("Case not implemented, something's wrong here.");
	}	
}