/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd.provenance;

import edu.freiburg.dbis.rdd.sparql.IRI;


/**
 * Provenance information for a constraint, specifiying the RDD position from
 * which a constraint origins.
 *
 * @author MSchmidt
 */
public class ConstraintProvenanceInfo {
	
	public enum ConstraintClass {
		CLASSES_CWA_CONSTRAINT,			// CLASSES CWA
		CLASS_CWA_CONSTRAINT,			// CLASS CWA
		CLASS_SINGLETON_CONSTRAINT,		// CLASS SINGLETON
		CLASS_KEY_CONSTRAINT,			// KEYs		
		QUALIFIED_PROPERTY_CONSTRAINT,	// QUALIFIED PROP CONSTRAINTS
		PROPERTIES_CWA_CONSTRAINT,		// PROPERTIES CWA
		UNQUALIFIED_PROPERTY_CONSTRAINT	// UNQUALIFIED PROP CONSTRAINTS
	}
	
	enum PropConstraintClass {
		DOMAIN_CONSTRAINT("DomainConstraint"),
		MIN_CONSTRAINT("MinConstraint"),
		MAX_CONSTRAINT("MaxConstraint"),
		PARTIALITY_CONSTRAINT("PartialityConstraint"),
		PATH_CONSTRAINT("PathConstraint"),
		RANGE_CONSTRAINT("RangeConstraint"),
		RANGE_TYPE_CONSTRAINT("RangeTypeConstraint"),
		SUBPROPERTY_CONSTRAINT("SubPropertyConstraint"),
		TOTALITY_CONSTRAINT("TotalityConstraint");
		
		PropConstraintClass(String constraintClassName) {
			this.constraintClassName = constraintClassName;
		}
		
		private String constraintClassName;
		
		@Override
		public String toString() {
			return constraintClassName;
		}
	}
	

	private int id; // for clustering of constraints sharing the same prov. info
	private String constraintRep;	
	private ConstraintClass constraintClass;
	private IRI referredClass;
	private PropConstraintClass propConstraintClass;
	
	public ConstraintProvenanceInfo(String constraintRep,
			ConstraintClass constraintClass, IRI referredClass,
			PropConstraintClass propConstraintClass) {
		this.constraintRep= constraintRep;
		this.id = IdCounter.getInstance().freshId();
		this.constraintClass = constraintClass;
		this.referredClass = referredClass;
		this.propConstraintClass = propConstraintClass;
	}
	
	public int getId() {
		return id;
	}
	
	public ConstraintClass getConstraintClass() {
		return constraintClass;
	}
	
	public IRI getReferredClass() {
		return referredClass;
	}
	
	public PropConstraintClass getPropConstraintClass() {
		return propConstraintClass;
	}
	
	public String getConstraintRep() {
		return constraintRep;
	}
	
	public String constraintClassAsString() {

		switch (constraintClass) {
		case CLASSES_CWA_CONSTRAINT:
			return "[CLASSES CWA]";
		case CLASS_CWA_CONSTRAINT:
			return "[CLASS CWA]";
		case CLASS_SINGLETON_CONSTRAINT:
			return "[CLASS SINGLETON]";
		case CLASS_KEY_CONSTRAINT:
			return "[Key]";
		case QUALIFIED_PROPERTY_CONSTRAINT:
			return "[Qualified " + propConstraintClass + "]";
		case PROPERTIES_CWA_CONSTRAINT:
			return "[PROPERTIES CWA]";
		case UNQUALIFIED_PROPERTY_CONSTRAINT:
			return "[Unqualified " + propConstraintClass + "]";
		default:
			throw new RuntimeException("Illegal constraint type: " + constraintClass);
		}
	}
	
	@Override
	public String toString() {
		StringBuilder buf = new StringBuilder();
		buf.append("#### " + constraintClassAsString() + "\n");
		buf.append("## " + constraintRep);
		return buf.toString();
	}
	
	@Override
	public int hashCode() {
		return getId();
	}
	
	/**
	 * Define equality over the ID
	 */
	@Override
	public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (obj == this)
            return true;
        if (!(obj instanceof ConstraintProvenanceInfo))
            return false;

        return this.hashCode()==obj.hashCode();
	}
}