/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd.parser;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.List;

import org.junit.Test;

import edu.freiburg.dbis.rdd.model.ClassConstraint;
import edu.freiburg.dbis.rdd.model.DomainConstraint;
import edu.freiburg.dbis.rdd.model.IRIWithRangeType;
import edu.freiburg.dbis.rdd.model.Key;
import edu.freiburg.dbis.rdd.model.MaxConstraint;
import edu.freiburg.dbis.rdd.model.MinConstraint;
import edu.freiburg.dbis.rdd.model.PartialityConstraint;
import edu.freiburg.dbis.rdd.model.PathConstraint;
import edu.freiburg.dbis.rdd.model.PropConstraint;
import edu.freiburg.dbis.rdd.model.PropConstraintSec;
import edu.freiburg.dbis.rdd.model.RDD;
import edu.freiburg.dbis.rdd.model.RangeConstraint;
import edu.freiburg.dbis.rdd.model.RangeTypeBNode;
import edu.freiburg.dbis.rdd.model.RangeTypeIRI;
import edu.freiburg.dbis.rdd.model.RangeTypeLiteral;
import edu.freiburg.dbis.rdd.model.RangeTypeResource;
import edu.freiburg.dbis.rdd.model.SubPropertyConstraint;
import edu.freiburg.dbis.rdd.model.TotalityConstraint;

/**
 * Comprehensive test covering all the language constructs
 * in combination.
 * 
 * @author MSchmidt
 */
public class MixedFeaturesTest {

	private static String TEST_BASE_DIR = "examples/test/parse/";

	
	@Test
	public void testMixedFeaturesParsing()
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "MixedFeatures.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	
	    
	    System.out.println(rdd);

	    ///// VERIFY CLASS CONSTRAINT SECTION
	    assertFalse(rdd.getClassConstraintSec().isOWA());
	    List<ClassConstraint> ccs = rdd.getClassConstraintSec().getClassConstraints();
	    assertEquals(3,ccs.size());
	    
	    // class 1 "header"
	    // OWA CLASS :c1 
	    ClassConstraint cc1 = ccs.get(0);
	    assertEquals("<http://my.default.de/c1>",cc1.getClazz().toString());
	    assertTrue(cc1.isOWA());
	    assertFalse(cc1.isSingleton());
	    assertTrue(cc1.getSubClazzes().isEmpty());
	    
	    // class 2 "header"
	    // "CWA SINGLETON CLASS <http://my.default.de/c2>"
	    ClassConstraint cc2 = ccs.get(1);
	    assertEquals("<http://my.default.de/c2>",cc2.getClazz().toString());
	    assertFalse(cc2.isOWA());
	    assertTrue(cc2.isSingleton());
	    assertTrue(cc2.getSubClazzes().isEmpty());
	    
	    // class 3 "header"
	    // "CWA CLASS :c3  SUBCLASS 
	    //     <http://www.example.com/sc1>, example:sc2" 
	    ClassConstraint cc3 = ccs.get(2);
	    assertEquals("<http://my.default.de/c3>",cc3.getClazz().toString());
	    assertFalse(cc3.isOWA());
	    assertFalse(cc3.isSingleton());
	    assertEquals(2,cc3.getSubClazzes().size());
	    assertEquals("<http://www.example.com/sc1>",cc3.getSubClazzes().get(0).toString());
	    assertEquals("<http://www.example.com/sc2>",cc3.getSubClazzes().get(1).toString());
	    
	    // class 1-3 prop constraints (they are all identical,
	    // just in different syntaxes, i.e. prefix usage)
	    for (ClassConstraint cc : ccs) {
	    	List<PropConstraint> pcs = cc.getPropConstraints();

		    List<Key> keys = cc.getKeys();
		    assertEquals(2,keys.size());
		    
		    // KEY <http://www.example.com/p1> : IRI;
		    Key key1 = keys.get(0);
		    List<IRIWithRangeType> key1Props = key1.getKeyProps();
		    assertEquals(1, key1Props.size());
		    assertEquals("<http://www.example.com/p1>", key1Props.get(0).getIRI().toString());
		    assertTrue(key1Props.get(0).getRangeType() instanceof RangeTypeIRI);
		    
		    // KEY <http://www.example.com/p2> : IRI, <http://www.example.com/p3> : RESOURCE; 
		    Key key2 = keys.get(1);
		    List<IRIWithRangeType> key2Props = key2.getKeyProps();
		    assertEquals(2, key2Props.size());
		    assertEquals("<http://www.example.com/p2>", key2Props.get(0).getIRI().toString());
		    assertEquals("<http://www.example.com/p3>", key2Props.get(1).getIRI().toString());
		    assertTrue(key2Props.get(0).getRangeType() instanceof RangeTypeIRI);
		    assertTrue(key2Props.get(1).getRangeType() instanceof RangeTypeResource);
		    
	    	// properties are identical throughout all classes
	    	assertProperties(pcs);
	    }
	    
	    ///// VERIFY PROP CONSTRAINT SECTION
	    PropConstraintSec pcs = rdd.getPropConstraintSec();
	    assertTrue(pcs.isOWA());
	    assertProperties(pcs.getPropConstraints());
	}


	/**
	 * Asserts that the properties are like this:
	 * 
	 * <http://www.example.com/p1> : LITERAL(<http://my.datatypes#integer>);
     * DOMAIN(<http://www.example.com/C>) <http://www.example.com/p2> : LITERAL;
     * RANGE(<http://www.example.com/C>) <http://www.example.com/p3>;
     * MAX(2) <http://www.example.com/p4> : IRI;
     * MIN(2) <http://www.example.com/p5>;
     * PARTIAL <http://www.example.com/p6> : BNODE;
     * TOTAL <http://www.example.com/p7>;
     * PATH(<http://www.example.com/u1>/<http://www.example.com/u2>) <http://www.example.com/p8> : RESOURCE;
     * SUBPROPERTY(<http://www.example.com/u1>,<http://www.example.com/u2>) <http://www.example.com/p9>;
	 */
	private void assertProperties(List<PropConstraint> pcs) {
		assertEquals(9,pcs.size());
		
		PropConstraint c1 = pcs.get(0);
		assertEquals("<http://www.example.com/p1>", c1.getPropWithRange().getIRI().toString());
		assertTrue(c1.getPropWithRange().getRangeType() instanceof RangeTypeLiteral);
		assertEquals("<http://my.datatypes#integer>", ((RangeTypeLiteral)c1.getPropWithRange().getRangeType()).getDatatype().toSPARQL());

		DomainConstraint c2 = (DomainConstraint)pcs.get(1);
		assertEquals("<http://www.example.com/p2>", c2.getPropWithRange().getIRI().toString());
		assertEquals("<http://www.example.com/C>", c2.getDomain().toString());
		assertTrue(c2.getPropWithRange().getRangeType() instanceof RangeTypeLiteral);
		assertNull(((RangeTypeLiteral)c2.getPropWithRange().getRangeType()).getDatatype());
		
		RangeConstraint c3 = (RangeConstraint)pcs.get(2);
		assertEquals("<http://www.example.com/p3>", c3.getPropWithRange().getIRI().toString());
		assertEquals("<http://www.example.com/C>", c3.getRange().toString());
		assertNull(c3.getPropWithRange().getRangeType());
		
		MaxConstraint c4 = (MaxConstraint)pcs.get(3);
		assertEquals("<http://www.example.com/p4>", c4.getPropWithRange().getIRI().toString());
		assertEquals(2, c4.getNr());
		assertTrue(c4.getPropWithRange().getRangeType() instanceof RangeTypeIRI);

		MinConstraint c5 = (MinConstraint)pcs.get(4);
		assertEquals("<http://www.example.com/p5>", c5.getPropWithRange().getIRI().toString());
		assertEquals(2, c5.getNr());
		assertNull(c5.getPropWithRange().getRangeType());

		PartialityConstraint c6 = (PartialityConstraint)pcs.get(5);
		assertEquals("<http://www.example.com/p6>", c6.getPropWithRange().getIRI().toString());
		assertTrue(c6.getPropWithRange().getRangeType() instanceof RangeTypeBNode);
		
		TotalityConstraint c7 = (TotalityConstraint)pcs.get(6);
		assertEquals("<http://www.example.com/p7>", c7.getPropWithRange().getIRI().toString());
		assertNull(c7.getPropWithRange().getRangeType());

		PathConstraint c8 = (PathConstraint)pcs.get(7);
		assertEquals("<http://www.example.com/p8>", c8.getPropWithRange().getIRI().toString());
		assertEquals(2,c8.getPath().size());
		assertEquals("<http://www.example.com/u1>", c8.getPath().get(0).toString());
		assertEquals("<http://www.example.com/u2>", c8.getPath().get(1).toString());
		
		SubPropertyConstraint c9 = (SubPropertyConstraint)pcs.get(8);
		assertEquals("<http://www.example.com/p9>", c9.getPropWithRange().getIRI().toString());
		assertEquals(2, c9.getSubProps().size());
		assertEquals("<http://www.example.com/u1>", c9.getSubProps().get(0).toString());
		assertEquals("<http://www.example.com/u2>", c9.getSubProps().get(1).toString());
	}	
}
