/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd.parser;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.LinkedList;
import java.util.List;

import org.junit.Test;

import edu.freiburg.dbis.rdd.model.ClassConstraint;
import edu.freiburg.dbis.rdd.model.ClassConstraintSec;
import edu.freiburg.dbis.rdd.model.DomainConstraint;
import edu.freiburg.dbis.rdd.model.IRIWithRangeType;
import edu.freiburg.dbis.rdd.model.MaxConstraint;
import edu.freiburg.dbis.rdd.model.MinConstraint;
import edu.freiburg.dbis.rdd.model.PropConstraint;
import edu.freiburg.dbis.rdd.model.PropConstraintSec;
import edu.freiburg.dbis.rdd.model.RDD;
import edu.freiburg.dbis.rdd.model.RangeConstraint;
import edu.freiburg.dbis.rdd.model.RangeTypeBNode;
import edu.freiburg.dbis.rdd.model.RangeTypeIRI;
import edu.freiburg.dbis.rdd.model.RangeTypeLiteral;
import edu.freiburg.dbis.rdd.model.RangeTypeResource;

/**
 * Tests for OWA/CWA + SINGLETON keywords and the like.
 * 
 * @author MSchmidt
 */
public class MixedKeywordParsingTest {

	private static String TEST_BASE_DIR = "examples/test/parse/";
	
	@Test
	public void testOWACWASingletonParsing()
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "OWACWASingleton.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	
	    
	    ClassConstraintSec ccs = rdd.getClassConstraintSec();
	    PropConstraintSec pcs = rdd.getPropConstraintSec();
	    
	    for (ClassConstraint cc : ccs.getClassConstraints()) {
	    	if (cc.getClazz().toString().equals("<http://www.example.com/c1>"))
	    		assertTrue(cc.isOWA() && !cc.isSingleton());
	    	else if (cc.getClazz().toString().equals("<http://www.example.com/c2>"))
	    		assertTrue(!cc.isOWA() && !cc.isSingleton());
	    	else if (cc.getClazz().toString().equals("<http://www.example.com/c3>"))
	    		assertTrue(cc.isOWA() && cc.isSingleton());
	    	else if (cc.getClazz().toString().equals("<http://www.example.com/c4>"))
	    		assertTrue(!cc.isOWA() && cc.isSingleton());
	    	else
	    		assertTrue(false); // this case should never happen
	    }
	    assertTrue(pcs.isOWA());
	}
	
	@Test
	public void testAllRangeTypeParsing()
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "AllRangeTypes.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	
	    
	    ClassConstraintSec ccs = rdd.getClassConstraintSec();
	    // collect constraints
	    List<PropConstraint> qpcs = new LinkedList<PropConstraint>();
	    for (ClassConstraint cc : ccs.getClassConstraints())
	    	qpcs.addAll(cc.getPropConstraints());
	    assertAllRangeTypesPresent(qpcs);
	    
	    PropConstraintSec pcs = rdd.getPropConstraintSec();
	    assertAllRangeTypesPresent(pcs.getPropConstraints());
	}

	@Test
	public void testCapitalizationParsing()
	throws FileNotFoundException, ParseException {

		// the file is the same as for testAllRangeTypes,
		// but keywords in upper/lower case mix
		String rddFile = TEST_BASE_DIR + "Capitalization.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	
	    
	    ClassConstraintSec ccs = rdd.getClassConstraintSec();
	    PropConstraintSec pcs = rdd.getPropConstraintSec();

	    assertFalse(ccs.isOWA());
	    assertTrue(pcs.isOWA());
	    
	    for (ClassConstraint cc : ccs.getClassConstraints()) {
	    	assertTrue(cc.isOWA());
	    }
	    
	    // assert range types
	    List<PropConstraint> qpcs = new LinkedList<PropConstraint>();
	    for (ClassConstraint cc : ccs.getClassConstraints())
	    	qpcs.addAll(cc.getPropConstraints());
	    assertAllRangeTypesPresent(qpcs);	    
	    assertAllRangeTypesPresent(pcs.getPropConstraints());
	
	    // make sure the constraints were all recognized properly
	    assertTrue(Util.containsConstraint(ccs, PropConstraint.class));
	    assertTrue(Util.containsConstraint(ccs, MaxConstraint.class));
		assertTrue(Util.containsConstraint(ccs, MinConstraint.class));
		assertTrue(Util.containsConstraint(ccs, DomainConstraint.class));
		assertTrue(Util.containsConstraint(ccs, RangeConstraint.class));

	    assertTrue(Util.containsConstraint(pcs, PropConstraint.class));
	    assertTrue(Util.containsConstraint(pcs, MaxConstraint.class));
		assertTrue(Util.containsConstraint(pcs, MinConstraint.class));
		assertTrue(Util.containsConstraint(pcs, DomainConstraint.class));
		assertTrue(Util.containsConstraint(pcs, RangeConstraint.class));
	}

	
	private void assertAllRangeTypesPresent(List<PropConstraint> pcs) {
		boolean p1IsIRI = false;
		boolean p2IsBNODE = false;
		boolean p3IsRESOURCE = false;
		boolean p4IsLITERAL = false;
		boolean p5IsDTLITERAL = false;

		assertEquals(5,pcs.size());
		for (PropConstraint pc : pcs) {
			IRIWithRangeType iwrt = pc.getPropWithRange();

	    	if (iwrt.getIRI().toString().equals("<http://www.example.com/p1>"))
	    		p1IsIRI = iwrt.getRangeType().getClass().equals(RangeTypeIRI.class);
	    	else if (iwrt.getIRI().toString().equals("<http://www.example.com/p2>"))
	    		p2IsBNODE = iwrt.getRangeType().getClass().equals(RangeTypeBNode.class);
	    	else if (iwrt.getIRI().toString().equals("<http://www.example.com/p3>"))
	    		p3IsRESOURCE = iwrt.getRangeType().getClass().equals(RangeTypeResource.class);
	    	else if (iwrt.getIRI().toString().equals("<http://www.example.com/p4>"))
	    		p4IsLITERAL = iwrt.getRangeType().getClass().equals(RangeTypeLiteral.class);
	    	else if (iwrt.getIRI().toString().equals("<http://www.example.com/p5>"))
	    		p5IsDTLITERAL = iwrt.getRangeType().getClass().equals(RangeTypeLiteral.class) && ((RangeTypeLiteral)(iwrt.getRangeType())).getDatatype().toSPARQL().equals("<http://my.datatype>");	
		}
		
		assertTrue(p1IsIRI && p2IsBNODE && p3IsRESOURCE && p4IsLITERAL && p5IsDTLITERAL);
	}
	
	@Test
	public void testSubclassesParsing()
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "Subclasses.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	
	    
	    ClassConstraintSec ccs = rdd.getClassConstraintSec();
	    assertEquals(3,ccs.getClassConstraints().size());
	    
	    boolean c1NoSubclasses = false;
	    boolean c2OneSubclass = false;
	    boolean c3ThreeSubclasses = false;
	    for (ClassConstraint cc : ccs.getClassConstraints()) {
	    	if (cc.getClazz().toString().equals("<http://www.example.com/c1>"))
	    		c1NoSubclasses = cc.getSubClazzes().isEmpty();
	    	else if (cc.getClazz().toString().equals("<http://www.example.com/c2>"))
	    		c2OneSubclass = 
	    			cc.getSubClazzes().size()==1 &&
	    			cc.getSubClazzes().get(0).toString().equals("<http://www.example.com/sc1>");
	    	else if (cc.getClazz().toString().equals("<http://www.example.com/c3>")) {
	    		c3ThreeSubclasses = 
		    			cc.getSubClazzes().size()==3 &&
		    			cc.getSubClazzes().get(0).toString().equals("<http://www.example.com/sc1>") &&
	    				cc.getSubClazzes().get(1).toString().equals("<http://www.example.com/sc2>") &&
	    				cc.getSubClazzes().get(2).toString().equals("<http://www.example.com/sc3>");
	    	}
	    	else
	    		assertTrue(false); // must not happen
	    }
	    assertTrue(c1NoSubclasses && c2OneSubclass && c3ThreeSubclasses);
	}

}
