/*
Copyright 2014, Albert-Ludwigs-Universitt Freiburg (Database Group)

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
package edu.freiburg.dbis.rdd.semantics;

import static org.junit.Assert.assertEquals;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.util.Set;

import org.junit.Test;

import edu.freiburg.dbis.rdd.fol.FOFormula;
import edu.freiburg.dbis.rdd.model.RDD;
import edu.freiburg.dbis.rdd.parser.ParseException;
import edu.freiburg.dbis.rdd.parser.RDDParser;

/**
 * Test the semantics of range type specifications,
 * also in interaction with other constraints.
 * 
 * @author MSchmidt
 */
public class RangeTypeTest {

	protected static String TEST_BASE_DIR = "examples/test/semantics/";

	@Test
	public void testRangeTypeInclusion() 
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "RangeType.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	

	    Set<FOFormula> formulas = rdd.toFOL();

	    /**
	     * OWA CLASSES {
	     *   OWA CLASS <http://www.example.com/C1> {
	     *     <http://www.example.com/p1> : LITERAL(<http://www.example.com/INT>);
	     *     MIN(1) <http://www.example.com/p2> : RESOURCE;
	     *   }
	     * }
	     * 
	     * OWA PROPERTIES {
	     *   <http://www.example.com/p3> : LITERAL;
	     *   MAX(1) <http://www.example.com/p4> : BNODE;
	     *   DOMAIN(<http://www.example.com/C2>) <http://www.example.com/p5> : IRI;
	     * }
	     */
	    
	    for (FOFormula formula : formulas) { 
	    	System.err.println(formula.toRIF());
	    }
	    // We expect 8 constraints:
	    // - 3 non-range type (1 for C1, 2 for C2)
	    // - 5 range type
	    assertEquals(8,formulas.size());
	}

	@Test
	public void testRangeTypeInclusionInKey() 
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "RangeTypeInKey.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	

	    Set<FOFormula> formulas = rdd.toFOL();

	    /**
	     * OWA CLASSES {
	     *   OWA CLASS <http://www.example.com/C1> {
	     *     KEY <http://www.example.com/p1> : LITERAL, <http://www.example.com/p2>, <http://www.example.com/p3> : BNODE;
	     *   }
	     * }
	     */
	    // we expect the following constraints
	    // - 3 MIN(1) constraints
	    // - 3 MAX(1) constraints
	    // - the key constraint itself
	    // - 2 range type constraints
	    // => 9 in total
	    assertEquals(9,formulas.size());
	}

	@Test
	public void testQualifiedRangeTypeVariants() 
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "QualifiedRangeTypeVariants.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	

	    Set<FOFormula> formulas = rdd.toFOL();

	    /**
	     * OWA CLASSES {
	     *   OWA CLASS <http://www.example.com/C1> {
	     *     <http://www.example.com/p1> : IRI;
	     *     <http://www.example.com/p2> : RESOURCE;
	     *     <http://www.example.com/p3> : BNODE;
	     *     <http://www.example.com/p4> : LITERAL;
	     *     <http://www.example.com/p5> : LITERAL(<http://www.example.com/INT>);
	     *   }
	     * }
	     */
	    String rtIRI = 
	    	Util.serializedRangeTypeConstraint(Util.CLASS1, Util.PROP1, "IRI", null);
	    String rtResource = 
		    Util.serializedRangeTypeConstraint(Util.CLASS1, Util.PROP2, "RESOURCE", null);
	    String rtBNode = 
			Util.serializedRangeTypeConstraint(Util.CLASS1, Util.PROP3, "BNODE", null);
	    String rtLiteral =
	    	Util.serializedRangeTypeConstraint(Util.CLASS1, Util.PROP4, "LITERAL", null);
	    String rtTypedLiteral =
	    	Util.serializedRangeTypeConstraint(Util.CLASS1, Util.PROP5, "LITERAL", Util.INTEGER);
	    
	    Util.assertSemantics(formulas,
	    	rtIRI, rtResource, rtBNode, rtLiteral, rtTypedLiteral);
	}
	
	@Test
	public void testUnqualifiedRangeTypeVariants() 
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "UnqualifiedRangeTypeVariants.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	

	    Set<FOFormula> formulas = rdd.toFOL();

	    /**
	     * OWA PROPERTIES {
	     *   <http://www.example.com/p1> : IRI;
	     *   <http://www.example.com/p2> : RESOURCE;
	     *   <http://www.example.com/p3> : BNODE;
	     *   <http://www.example.com/p4> : LITERAL;
	     *   <http://www.example.com/p5> : LITERAL(<http://www.example.com/INT>);
	     * }
	     */
	    String rtIRI = 
		    Util.serializedRangeTypeConstraint(null, Util.PROP1, "IRI", null);
	    String rtResource = 
			Util.serializedRangeTypeConstraint(null, Util.PROP2, "RESOURCE", null);
	    String rtBNode = 
			Util.serializedRangeTypeConstraint(null, Util.PROP3, "BNODE", null);
	    String rtLiteral =
			Util.serializedRangeTypeConstraint(null, Util.PROP4, "LITERAL", null);
	    String rtTypedLiteral =
			Util.serializedRangeTypeConstraint(null, Util.PROP5, "LITERAL", Util.INTEGER);
	    
	    Util.assertSemantics(formulas,
	    	rtIRI, rtResource, rtBNode, rtLiteral, rtTypedLiteral);
	}
	
	@Test
	public void testCombinedConstraintAndRangeType() 
	throws FileNotFoundException, ParseException {
		
		String rddFile = TEST_BASE_DIR + "CombinedConstraintAndRangeType.rdd";
		FileInputStream fis = new FileInputStream(new File(rddFile));

	    RDDParser parser = new RDDParser(fis);
	    RDD rdd = parser.rdd();	

	    Set<FOFormula> formulas = rdd.toFOL();

	    /**
	     * OWA CLASSES {
	     *   OWA CLASS <http://www.example.com/C1> {
	     *     MIN(1) <http://www.example.com/p1> : IRI;
	     *   }
	     * }
	     * 
	     * OWA PROPERTIES {
	     *   MIN(1) <http://www.example.com/p2> : IRI;
	     * }
	     */
	    String qualifiedMin = 
	    	Util.serializedMin1Constraint(Util.CLASS1, Util.PROP1);
	    String qualifiedRt =
		    Util.serializedRangeTypeConstraint(Util.CLASS1, Util.PROP1, "IRI", null);

	    String unqualifiedMin =
		    Util.serializedMin1Constraint(null, Util.PROP2);
	    String unqualifiedRt= 
			Util.serializedRangeTypeConstraint(null, Util.PROP2, "IRI", null);
	    
	    Util.assertSemantics(formulas,
	    	qualifiedMin, qualifiedRt, unqualifiedMin, unqualifiedRt);
	}	
}